@@ -144,7 +144,18 @@ class InvalidColumns(Exception):
     pass
 
 
+_COUNTS_TABLE_CREATE_SQL = """
+CREATE TABLE IF NOT EXISTS [{}](
+   [table] TEXT PRIMARY KEY,
+   count INTEGER DEFAULT 0
+);
+""".strip()
+
+
 class Database:
+    _counts_table_name = "_counts"
+    use_counts_table = False
+
     def __init__(
         self,
         filename_or_conn=None,
@@ -152,6 +163,7 @@ class Database:
         recreate=False,
         recursive_triggers=True,
         tracer=None,
+        use_counts_table=False,
     ):
         assert (filename_or_conn is not None and not memory) or (
             filename_or_conn is None and memory
@@ -169,6 +181,7 @@ class Database:
         if recursive_triggers:
             self.execute("PRAGMA recursive_triggers=on;")
         self._registered_functions = set()
+        self.use_counts_table = use_counts_table
 
     @contextlib.contextmanager
     def tracer(self, tracer=None):
@@ -223,7 +236,7 @@ class Database:
         klass = View if table_name in self.view_names() else Table
         return klass(self, table_name, **kwargs)
 
-    def escape(self, value):
+    def quote(self, value):
         # Normally we would use .execute(sql, [params]) for escaping, but
         # occasionally that isn't available - most notable when we need
         # to include a "... DEFAULT 'value'" in a column definition.
@@ -267,6 +280,11 @@ class Database:
             ).fetchall()
         ]
 
+    @property
+    def triggers_dict(self):
+        "Returns {trigger_name: sql} dictionary"
+        return {trigger.name: trigger.sql for trigger in self.triggers}
+
     @property
     def journal_mode(self):
         return self.execute("PRAGMA journal_mode;").fetchone()[0]
@@ -279,6 +297,40 @@ class Database:
         if self.journal_mode != "delete":
             self.execute("PRAGMA journal_mode=delete;")
 
+    def _ensure_counts_table(self):
+        with self.conn:
+            self.execute(_COUNTS_TABLE_CREATE_SQL.format(self._counts_table_name))
+
+    def enable_counts(self):
+        self._ensure_counts_table()
+        for table in self.tables:
+            if (
+                table.virtual_table_using is None
+                and table.name != self._counts_table_name
+            ):
+                table.enable_counts()
+        self.use_counts_table = True
+
+    def cached_counts(self, tables=None):
+        sql = "select [table], count from {}".format(self._counts_table_name)
+        if tables:
+            sql += " where [table] in ({})".format(", ".join("?" for table in tables))
+        try:
+            return {r[0]: r[1] for r in self.execute(sql, tables).fetchall()}
+        except OperationalError:
+            return {}
+
+    def reset_counts(self):
+        tables = [table for table in self.tables if table.has_counts_triggers]
+        with self.conn:
+            self._ensure_counts_table()
+            counts_table = self[self._counts_table_name]
+            counts_table.delete_where()
+            counts_table.insert_all(
+                {"table": table.name, "count": table.execute_count()}
+                for table in tables
+            )
+
     def execute_returning_dicts(self, sql, params=None):
         cursor = self.execute(sql, params or tuple())
         keys = [d[0] for d in cursor.description]
@@ -399,7 +451,7 @@ class Database:
                 column_extras.append("NOT NULL")
             if column_name in defaults and defaults[column_name] is not None:
                 column_extras.append(
-                    "DEFAULT {}".format(self.escape(defaults[column_name]))
+                    "DEFAULT {}".format(self.quote(defaults[column_name]))
                 )
             if column_name in foreign_keys_by_column:
                 column_extras.append(
@@ -579,12 +631,15 @@ class Queryable:
         self.db = db
         self.name = name
 
-    @property
-    def count(self):
+    def execute_count(self):
         return self.db.execute(
             "select count(*) from [{}]".format(self.name)
         ).fetchone()[0]
 
+    @property
+    def count(self):
+        return self.execute_count()
+
     @property
     def rows(self):
         return self.rows_where()
@@ -668,6 +723,14 @@ class Table(Queryable):
             else " ({})".format(", ".join(c.name for c in self.columns)),
         )
 
+    @property
+    def count(self):
+        if self.db.use_counts_table:
+            counts = self.db.cached_counts([self.name])
+            if counts:
+                return next(iter(counts.values()))
+        return self.execute_count()
+
     def exists(self):
         return self.name in self.db.table_names()
 
@@ -759,6 +822,11 @@ class Table(Queryable):
             ).fetchall()
         ]
 
+    @property
+    def triggers_dict(self):
+        "Returns {trigger_name: sql} dictionary"
+        return {trigger.name: trigger.sql for trigger in self.triggers}
+
     def create(
         self,
         columns,
@@ -1084,9 +1152,7 @@ class Table(Queryable):
             col_type = str
         not_null_sql = None
         if not_null_default is not None:
-            not_null_sql = "NOT NULL DEFAULT {}".format(
-                self.db.escape(not_null_default)
-            )
+            not_null_sql = "NOT NULL DEFAULT {}".format(self.db.quote(not_null_default))
         sql = "ALTER TABLE [{table}] ADD COLUMN [{col_name}] {col_type}{not_null_default};".format(
             table=self.name,
             col_name=col_name,
@@ -1169,6 +1235,60 @@ class Table(Queryable):
         self.db.add_foreign_keys([(self.name, column, other_table, other_column)])
         return self
 
+    def enable_counts(self):
+        sql = (
+            textwrap.dedent(
+                """
+        {create_counts_table}
+        CREATE TRIGGER IF NOT EXISTS [{table}{counts_table}_insert] AFTER INSERT ON [{table}]
+        BEGIN
+            INSERT OR REPLACE INTO [{counts_table}]
+            VALUES (
+                {table_quoted},
+                COALESCE(
+                    (SELECT count FROM [{counts_table}] WHERE [table] = {table_quoted}),
+                0
+                ) + 1
+            );
+        END;
+        CREATE TRIGGER IF NOT EXISTS [{table}{counts_table}_delete] AFTER DELETE ON [{table}]
+        BEGIN
+            INSERT OR REPLACE INTO [{counts_table}]
+            VALUES (
+                {table_quoted},
+                COALESCE(
+                    (SELECT count FROM [{counts_table}] WHERE [table] = {table_quoted}),
+                0
+                ) - 1
+            );
+        END;
+        INSERT OR REPLACE INTO _counts VALUES ({table_quoted}, (select count(*) from [{table}]));
+        """
+            )
+            .strip()
+            .format(
+                create_counts_table=_COUNTS_TABLE_CREATE_SQL.format(
+                    self.db._counts_table_name
+                ),
+                counts_table=self.db._counts_table_name,
+                table=self.name,
+                table_quoted=self.db.quote(self.name),
+            )
+        )
+        with self.db.conn:
+            self.db.conn.executescript(sql)
+        self.db.use_counts_table = True
+
+    @property
+    def has_counts_triggers(self):
+        trigger_names = {
+            "{table}{counts_table}_{suffix}".format(
+                counts_table=self.db._counts_table_name, table=self.name, suffix=suffix
+            )
+            for suffix in ["insert", "delete"]
+        }
+        return trigger_names.issubset(self.triggers_dict.keys())
+
     def enable_fts(
         self,
         columns,
@@ -1583,6 +1703,7 @@ class Table(Queryable):
         )
 
         with self.db.conn:
+            result = None
             for query, params in queries_and_params:
                 try:
                     result = self.db.execute(query, params)
